import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf

class AtrousCLSTM(tf.keras.Model):

    def create_atrous_conv_layer(self, num_filters, size, dilation):
        conv = tf.keras.layers.Conv1D(num_filters, size, strides=1, padding='valid',
                                      data_format='channels_last', dilation_rate=dilation, groups=1,
                                      activation="elu", kernel_initializer='glorot_uniform')
        return conv
    def create_lstm_layer(self, hidden_state_dimension,sequence=True, direction="unidirectional", name=""):
        if direction == "bidirectional":
            hidden_state_dimension = int(hidden_state_dimension / 2)
        lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                        recurrent_activation='sigmoid', return_sequences=sequence, return_state=False,name="forward_lstm_{}".format(name))

        if direction == "bidirectional":
            # already returns the reversed sequence
            bwd_lstm_layer = tf.keras.layers.LSTM(units=hidden_state_dimension,
                                                recurrent_activation='sigmoid', go_backwards=True,
                                                return_sequences=sequence, return_state=False, name="backward_lstm_{}".format(name))
            lstm_layer = tf.keras.layers.Bidirectional(layer=lstm_layer, merge_mode='concat',
                                                      backward_layer=bwd_lstm_layer, name="full_bi_lstm_{}".format(name))

        return lstm_layer

    @tf.function
    def wav_to_spectrogram(self, binary):
        # Convert the waveform to a spectrogram via a STFT.
        window = tf.constant(30)
        spectrogram = tf.signal.stft(
            binary, frame_length=window, frame_step=window)
        # Obtain the magnitude of the STFT.
        spectrogram = tf.abs(spectrogram)
        # Add a `channels` dimension, so that the spectrogram can be used
        # as image-like input data with convolution layers (which expect
        # shape (`batch_size`, `height`, `width`, `channels`).
        spectrogram = tf.expand_dims(spectrogram, axis=-1)
        return spectrogram

    @tf.function(experimental_relax_shapes=True)
    def call(self, recording, training=False):
        x = tf.cast(tf.expand_dims(recording, axis=-1), tf.float32)
        # atrous temporal CNN
        for i, l in enumerate(self.atrconv_list):
            x = l(x)
            x = self.pooling_list[i](x)
            x = self.bn_list[i](x, training=training)
        # temporal CNN
        for j, l in enumerate(self.conv_list[:-1]):
            x = l(x)
            x = self.pooling_list[i + 1 + j](x)
            x = self.bn_list[i + 1 + j](x, training=training)
        # last temporal CNN
        x = self.conv_list[-1](x)
        # biLSTMS
        for l in self.bi_lstm:
            x = l(x)
        x = self.drop(x, training=training)
        dense = self.dense(x)
        dense2 = self.dense2(dense)
        return dense2

    def __init__(self):
        super(AtrousCLSTM, self).__init__()

        self.atrconv_list = [self.create_atrous_conv_layer(256, 5, 2), self.create_atrous_conv_layer(256, 5, 2)]
        self.conv_list = [self.create_atrous_conv_layer(256, 5, 1) for i in range(4)]
        self.pooling_list = [tf.keras.layers.MaxPooling1D(pool_size=2, strides=2) for i in range(5)]
        self.bi_lstm = [self.create_lstm_layer(128, direction="bidirectional", sequence=False) for i in range(1)]
        self.bn_list = [tf.keras.layers.BatchNormalization() for i in range(5)]
        self.drop = tf.keras.layers.Dropout(0.5)
        self.dense = tf.keras.layers.Dense(100, activation="elu")
        self.dense2 = tf.keras.layers.Dense(3)
