#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
import preprocessing
import sven_model as sven_model
import os
import evaluate_model
os.environ['CUDA_VISIBLE_DEVICES']=""
os.environ['TF_CPP_MIN_LOG_LEVEL']='3'
import tensorflow as tf
import sklearn.model_selection
import multiprocessing as mp


################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################
batch_size = 50
epochs = 500
transfer_epochs = 50
length = 5000
gamma = 5
alpha = 3
# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)
    recordings, features, murmur_y, outcome_y = get_train_data(data_folder, verbose)
    auscul_locs = ['AV', 'MV', 'PV', 'TV', 'PhC']
    # dims {AV: [#subsamples, #time steps]}
    murmur_recs = dict()
    murmur_labels = dict()
    murmur_outcome = dict()
    murmur_features = dict()
    other_recs = dict()
    other_labels = dict()
    other_outcome = dict()
    other_features = dict()
    # sort recording according to murmur_locs:
    for i, lab in enumerate(murmur_y):
        if lab == 0:
            # murmur locations
            for ml in features[i][-1]:
                # all locations
                for j, l in enumerate(features[i][5]):
                    if ml == l:
                        if ml in murmur_recs:
                            murmur_recs[l].append(recordings[i][j].tolist())
                            murmur_labels[l].append(murmur_y[i])
                            murmur_outcome[l].append(outcome_y[i])
                            murmur_features[l].append(features[i][:4])
                        else:
                            murmur_recs[l] = [recordings[i][j].tolist()]
                            murmur_labels[l] = [murmur_y[i]]
                            murmur_outcome[l] = [outcome_y[i]]
                            murmur_features[l] = [features[i][:4]]
        else:
            for j, l in enumerate(features[i][5]):
                if l in other_recs:
                    other_recs[l].append(recordings[i][j].tolist())
                    other_labels[l].append(murmur_y[i])
                    other_outcome[l].append(outcome_y[i])
                    other_features[l].append(features[i][:4])
                else:
                    other_recs[l] = [recordings[i][j].tolist()]
                    other_labels[l] = [murmur_y[i]]
                    other_outcome[l] = [outcome_y[i]]
                    other_features[l] = [features[i][:4]]

    # generate tensorflow data sets:
    data_traindict = dict()
    data_valdict = dict()
    for loc in auscul_locs:
        recs = []
        labels = []
        outcomes = []
        features = []
        if loc in murmur_recs:
            recs += murmur_recs[loc]
            labels += murmur_labels[loc]
            outcomes += murmur_outcome[loc]
            features += murmur_features[loc]
        if loc in other_recs:
            recs += other_recs[loc]
            labels += other_labels[loc]
            outcomes += other_outcome[loc]
            features += other_features[loc]
        if len(recs) > 10:
            recs_train, recs_val,features_train, features_val, \
            labels_train, labels_val, outcomes_train, outcomes_val = sklearn.model_selection.train_test_split(
                recs, features, labels, outcomes, test_size=0.2, stratify=labels,
                random_state=42)
            data_traindict[loc] = (recs_train, features_train, labels_train, outcomes_train)
            data_valdict[loc] = (recs_val, features_val, labels_val, outcomes_val)

    # Train the model.
    if verbose >= 1:
        print('Training model...')
        # print(recordings[:2])

    # classes = ['Present', 'Unknown', 'Absent']
    # print(data_traindict["AV"][0][0])
    processes = list()
    for loc in data_traindict.keys():
        print(loc)
        train_location(loc, data_traindict[loc], data_valdict[loc], model_folder)
        #p = mp.get_context("spawn").Process(target=train_location, args=(
        #loc, data_traindict[loc], data_valdict[loc], model_folder))
        #processes.append(p)
        #p.start()
    #for process in processes:
    #    process.join()

    if verbose >= 1:
        print('Done.')


# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    auscul_locs = []
    for file in os.listdir(model_folder):
        d = os.path.join(model_folder, file)
        if os.path.isdir(d):
            auscul_locs.append(file)

    model_dict = {}
    for loc in auscul_locs:
        model = sven_model.Model()
        ckpt = tf.train.Checkpoint(step=tf.Variable(0), model=model)
        ckpt_manager = tf.train.CheckpointManager(ckpt, model_folder + "/" + str(loc),
                                                  max_to_keep=2)
        ckpt_path = ckpt_manager.latest_checkpoint
        print(ckpt_path)
        ckpt.restore(save_path=ckpt_path)
        model_dict[loc] = model
    return model_dict


# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome = len(outcome_classes)
    # Load features.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = 0.0

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_feature = 0
    if compare_strings(sex, 'Female'):
        sex_feature = 1
    elif compare_strings(sex, 'Male'):
        sex_feature = 2

    # Extract height and weight.
    height = get_height(data)
    if not height or np.isnan(height):
        height = 0.0
    weight = get_weight(data)
    if not weight or np.isnan(weight):
        weight = 0.0

    locs = list(map(lambda item: 'PhC' if item.lower() == 'phc' else item, get_locations(data)))
    murmur_sum_probs = np.array([0,0,0])
    outcome_sum_probs = np.array([0,0])
    count = 0
    # Murmur classification
    for i, loc in enumerate(locs):
        if loc not in model:
            continue
        freq = get_frequency(data)
        preprocessed_recording = preprocessing.preprocess_individual_recording(recordings[i], freq, length=length, harmonic = False, cycle = True)
        features = np.asarray([age, sex_feature, height, weight], dtype=np.float32)
        logits, logit_outcome = model[loc]([preprocessed_recording], [features])
        decision_murmur, loc_probabilities_murmur = get_murmur_decision(logits[0])
        murmur_sum_probs = np.add(murmur_sum_probs, loc_probabilities_murmur)
        count +=1
        # Outcome classification
        decision_outcome, loc_probabilities_outcome = get_outcome_decision(logit_outcome[0][0])
        outcome_sum_probs = np.add(outcome_sum_probs, loc_probabilities_outcome)

    # ensemble probability as computed by scikit random forests
    ensemble_prob_murmur = murmur_sum_probs / count
    max_murmur_index = np.argmax(ensemble_prob_murmur)
    res_murmur = np.array([1 if max_murmur_index == i else 0 for i in range(num_murmur)])

    # ensemble probability as computed by scikit random forests
    ensemble_prob_outcome = outcome_sum_probs / count
    max_outcome_index = np.argmax(ensemble_prob_outcome)
    res_outcome = np.array([1 if max_outcome_index == i else 0 for i in range(2)])

    tweak = [[0.46,0,0.25], None]
    if tweak[0] is not None:
        if ensemble_prob_murmur[0] >= tweak[0][0]:
            max_murmur_index = 0
        elif ensemble_prob_murmur[2] >= tweak[0][2]:
            max_murmur_index = 2
        else:
            max_murmur_index = np.argmax(ensemble_prob_murmur)
        res_murmur = [1 if max_murmur_index == i else 0 for i in range(3)]

    if tweak[1] is not None:
        if ensemble_prob_outcome[0] >= tweak[1][0]:
            max_outcome_index = 0
        else:
            max_outcome_index = np.argmax(ensemble_prob_outcome)
        res_outcome = [1 if max_outcome_index == i else 0 for i in range(2)]

    # Combination of two outputs
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((res_murmur, res_outcome))
    probabilities = np.concatenate((ensemble_prob_murmur, ensemble_prob_outcome))

    return classes, labels, probabilities


################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

def get_train_data(data_folder, verbose):
    # features:
    # [age, sex_feature, height, weight, is_pregnant, locations, frequency, segmentations, mal, mlocs]
    recordings, features, murmurs, outcomes = preprocessing.get_raw(data_folder, True, verbose)
    new_recs = list()
    for p, patient in enumerate(recordings):
        p_recs = list()
        for record in patient:
            last_cycle = preprocessing.preprocess_individual_recording(record, features[p][6], length=length, harmonic = False, cycle = True)
            p_recs.append(last_cycle)
        new_recs.append(p_recs)
    return new_recs, features, murmurs, outcomes




def train_location(loc, train, val, model_folder):
    # generate tensorflow datasets:
    def gen_tfdata(list_recordings, list_features, list_labels, list_outcomes):
        in_ds = tf.data.Dataset.zip(
            (tf.data.Dataset.from_tensor_slices(list_recordings),
             tf.data.Dataset.from_tensor_slices(list_features), tf.data.Dataset.from_tensor_slices(list_labels),
             tf.data.Dataset.from_tensor_slices(list_outcomes)))
        return in_ds.cache().shuffle(1000).batch(batch_size=batch_size).prefetch(tf.data.AUTOTUNE)

    tf_train = gen_tfdata(train[0], train[1], train[2], train[3])
    tf_valid = gen_tfdata(val[0], val[1], val[2], val[3])
    # generate model
    model = sven_model.Model()
    murmur_loss = sven_model.Categor_Focal_Loss(alpha=alpha, gamma=gamma, reduction="sum_over_batch_size")
    murmur_valid_loss = sven_model.Categor_Focal_Loss(alpha=alpha, gamma=gamma, reduction="none")
    murmur_opti = tf.optimizers.RMSprop(learning_rate=0.0001)
    outcome_loss = tf.losses.BinaryCrossentropy(from_logits=True)
    outcome_valid_loss = tf.losses.BinaryCrossentropy(from_logits=True, reduction=tf.keras.losses.Reduction.NONE)
    outcome_opti = tf.optimizers.RMSprop(learning_rate=0.001)

    def validate_murmur():
        murmur_classes = ['Present', 'Unknown', 'Absent']
        overall_validloss = np.array([])
        overall_labels = np.array([])
        overall_decision = np.array([])
        for recs, features, labels, outcomes in tf_valid:
            # loss computation
            # weights for "present"
            label_weights = tf.where(labels == 0, 5.0, tf.cast(labels, dtype=tf.dtypes.float32))
            # weights for "unknown"
            label_weights = tf.where(labels == 1, 3.0, label_weights)
            # weights for "absent"
            label_weights = tf.where(labels == 2, 1.0, label_weights)
            normalized_weights, _ = tf.linalg.normalize(label_weights, ord=1)
            normalized_weights = tf.size(labels, out_type=tf.dtypes.float32) * normalized_weights
            logits_murmur, logit_outcome = model(recs, features, training=False)
            overall_validloss = np.concatenate(
                (overall_validloss, murmur_valid_loss(labels, logits_murmur, sample_weight=normalized_weights).numpy()))
            # weighted average (murmur) computation
            overall_labels = np.concatenate((overall_labels, labels.numpy()))
            decisions, _ = get_murmur_decision(logits_murmur)
            overall_decision = np.concatenate((overall_decision, decisions))
        loss_valid = np.mean(overall_validloss)
        labels_one_hot = numpy_one_hot(overall_labels, 3)
        output_one_hot = numpy_one_hot(overall_decision, 3)
        murmur_score_valid = compute_weighted_accuracy(labels_one_hot,output_one_hot, murmur_classes)
        return murmur_score_valid, loss_valid, tf.reduce_sum(output_one_hot, axis=0)

    def validate_outcome():
        overall_validloss = np.array([])
        overall_outcomes = np.array([])
        overall_decision = np.array([])
        for recs, features, labels, outcomes in tf_valid:
            logits_murmur, logit_outcome = model(recs, features, training=False)
            overall_validloss = np.concatenate(
                (overall_validloss, outcome_valid_loss(tf.expand_dims(outcomes, axis=-1), logit_outcome).numpy()))
            # weighted average (murmur) computation
            overall_outcomes = np.concatenate((overall_outcomes, outcomes.numpy()))
            decisions = tf.where(tf.squeeze(tf.greater(logit_outcome, tf.constant(0.0)), axis=-1), x=tf.constant(1), y=tf.constant(0))
            overall_decision = np.concatenate((overall_decision, decisions))
        loss_valid = np.mean(overall_validloss)
        outcomes_one_hot = numpy_one_hot(overall_outcomes, 2)
        decisions_one_hot = numpy_one_hot(overall_decision, 2)
        #outcomes_named = tf.where(overall_outcomes > 0.5, 'Normal', 'Abnormal')
        #decisions_named = tf.where(overall_decision > 0.5, 'Normal', 'Abnormal')
        outcome_score_valid = compute_challenge_metric(outcomes_one_hot,decisions_one_hot, ['Abnormal', 'Normal'], ['Abnormal', 'Normal'])
        return outcome_score_valid, loss_valid


    best_valid = 0
    best_outcome_valid = 30000000
    ckpt = tf.train.Checkpoint(step=tf.Variable(0), model=model)
    ckpt_manager = tf.train.CheckpointManager(ckpt, model_folder + "/" + str(loc),
                                              max_to_keep=1)
    ckpt_manager.save(checkpoint_number=ckpt.step)
    for e in range(1, epochs + 1):
        for step, (recs, features, labels, outcomes) in enumerate(tf_train):
            # weights for "present"
            label_weights = tf.where(labels == 0, 5.0, tf.cast(labels, dtype=tf.dtypes.float32))
            # weights for "unknown"
            label_weights = tf.where(labels == 1, 3.0, label_weights)
            # weights for "absent"
            label_weights = tf.where(labels == 2, 1.0, label_weights)
            normalized_weights, _ = tf.linalg.normalize(label_weights, ord=1)
            normalized_weights = tf.size(labels, out_type=tf.dtypes.float32) * normalized_weights
            with tf.GradientTape() as tape:
                logits_murmur, logit_outcome = model(recs, features, training=True)
                current_loss = murmur_loss(labels, logits_murmur ,sample_weight=normalized_weights)
                #tf.print(model.trainable_variables[0].name)
                #tf.print(model.trainable_variables[-1].name)
                murmur_trainables = [var for var in model.trainable_variables if var.name.startswith("murmur_head")]
                gradient = tape.gradient(current_loss, murmur_trainables)
                #tf.print(gradient[0])
                murmur_opti.apply_gradients(zip(gradient, murmur_trainables))
        ckpt.step.assign_add(1)
        if int(ckpt.step) % 1 == 0:
            murmur_dec_val, res_valid, decisions = validate_murmur()
            if murmur_dec_val > best_valid:
                best_valid = murmur_dec_val
                ckpt_manager.save(checkpoint_number=ckpt.step)
                print("Current model weights stored for: {}".format(loc))
            print("Validation weighted accuracy / Loss / Decisions ({}) after {} epochs: {} / {} / {}".format(loc, e, murmur_dec_val, current_loss, decisions))

    ### epochs for the transfer learning part that only adapts the head for the outcome prediction
    ### 1) load best (for validation) model regarding murmur detection
    ### 2) train head for outcome prediction on top of that
    ckpt.restore(ckpt_manager.latest_checkpoint)
    for e in range(1, transfer_epochs + 1):
        for step, (recs, features, labels, outcomes) in enumerate(tf_train):
            with tf.GradientTape() as tape:
                # training=False to prevent dropout and adaptation of batch normalisation in murmur part
                logits_murmur, logit_outcome = model(recs, features, training=False)
                current_outcome_loss = outcome_loss(tf.expand_dims(outcomes, axis=-1), logit_outcome)
                outcome_trainables = [var for var in model.trainable_variables if var.name.startswith("outcome_head")]
                outcome_gradient = tape.gradient(current_outcome_loss, outcome_trainables)
                outcome_opti.apply_gradients(zip(outcome_gradient, outcome_trainables))
        ckpt.step.assign_add(1)
        outcome_score_val, res_valid = validate_outcome()
        if outcome_score_val < best_outcome_valid:
            best_outcome_valid = outcome_score_val
            ckpt_manager.save(checkpoint_number=ckpt.step)
            print("Current model weights (outcome head) stored for: {}".format(loc))
        print("Validation outcome score ({}) after {} epochs: {}".format(loc, e, outcome_score_val))
    tf.keras.backend.clear_session()
    return 1

def get_murmur_decision(logits):
    probabilities = tf.nn.softmax(logits).numpy()
    max_prob_index = np.argmax(probabilities, axis=-1)
    return max_prob_index, probabilities

def get_outcome_decision(logit):
    logit_np = logit.numpy()
    if logit_np >= 0:
        decision = 1
    else:
        decision = 0
    probabilities = [1-tf.sigmoid(logit).numpy(), tf.sigmoid(logit).numpy()]
    return decision, probabilities

def numpy_one_hot(a, max_ind):
    a = a.astype(np.int32)
    a_one_hot = np.eye(max_ind)[a]
    return a_one_hot

#### copied from the evaluate_model.py script as it might not be avaialble during training

# Compute a binary confusion matrix, where the columns are the expert labels and the rows are the classifier labels.
def compute_confusion_matrix(labels, outputs):
    assert(np.shape(labels)[0] == np.shape(outputs)[0])
    assert(all(value in (0, 1, True, False) for value in np.unique(labels)))
    assert(all(value in (0, 1, True, False) for value in np.unique(outputs)))

    num_patients = np.shape(labels)[0]
    num_label_classes = np.shape(labels)[1]
    num_output_classes = np.shape(outputs)[1]

    A = np.zeros((num_output_classes, num_label_classes))
    for k in range(num_patients):
        for i in range(num_output_classes):
            for j in range(num_label_classes):
                if outputs[k, i] == 1 and labels[k, j] == 1:
                    A[i, j] += 1

    return A

# Compute accuracy.
def compute_weighted_accuracy(labels, outputs, classes):
    # Define constants.
    if classes == ['Present', 'Unknown', 'Absent']:
        weights = np.array([[5, 3, 1], [5, 3, 1], [5, 3, 1]])
    elif classes == ['Abnormal', 'Normal']:
        weights = np.array([[5, 1], [5, 1]])
    else:
        raise NotImplementedError('Weighted accuracy undefined for classes {}'.format(', '.join(classes)))

    # Compute confusion matrix.
    assert(np.shape(labels) == np.shape(outputs))
    A = compute_confusion_matrix(labels, outputs)

    # Multiply the confusion matrix by the weight matrix.
    assert(np.shape(A) == np.shape(weights))
    B = weights * A

    # Compute weighted_accuracy.
    if np.sum(B) > 0:
        weighted_accuracy = np.trace(B) / np.sum(B)
    else:
        weighted_accuracy = float('nan')

    return weighted_accuracy

# Define total cost for algorithmic prescreening of m patients.
def cost_algorithm(m):
    return 10*m

# Define total cost for expert screening of m patients out of a total of n total patients.
def cost_expert(m, n):
    return (25 + 397*(m/n) -1718*(m/n)**2 + 11296*(m/n)**4) * n

# Define total cost for treatment of m patients.
def cost_treatment(m):
    return 10000*m

# Define total cost for missed/late treatement of m patients.
def cost_error(m):
    return 50000*m

# Compute Challenge metric.
def compute_challenge_metric(labels, outputs, label_classes, output_classes):
    # Define positive and negative classes for referral and treatment.
    positive_classes = ['Present', 'Unknown', 'Abnormal']
    negative_classes = ['Absent', 'Normal']

    # Compute confusion matrix.
    A = compute_confusion_matrix(labels, outputs)

    # Identify positive and negative classes for referral.
    idx_label_positive = [i for i, x in enumerate(label_classes) if x in positive_classes]
    idx_label_negative = [i for i, x in enumerate(label_classes) if x in negative_classes]
    idx_output_positive = [i for i, x in enumerate(output_classes) if x in positive_classes]
    idx_output_negative = [i for i, x in enumerate(output_classes) if x in negative_classes]

    # Identify true positives, false positives, false negatives, and true negatives.
    tp = np.sum(A[np.ix_(idx_output_positive, idx_label_positive)])
    fp = np.sum(A[np.ix_(idx_output_positive, idx_label_negative)])
    fn = np.sum(A[np.ix_(idx_output_negative, idx_label_positive)])
    tn = np.sum(A[np.ix_(idx_output_negative, idx_label_negative)])
    total_patients = tp + fp + fn + tn

    # Compute total cost for all patients.
    total_cost = cost_algorithm(total_patients) \
        + cost_expert(tp + fp, total_patients) \
        + cost_treatment(tp) \
        + cost_error(fn)

    # Compute mean cost per patient.
    if total_patients > 0:
        mean_cost = total_cost / total_patients
    else:
        mean_cost = float('nan')

    return mean_cost
