import sys
sys.path.insert(0, '..')
import helper_code as hc

import os
import shutil
import ntpath
from functools import partial
from multiprocessing.pool import ThreadPool
import sklearn.model_selection


def find_files(directory, id):
    filenames = list()
    for f in sorted(os.listdir(directory)):
        head, tail = ntpath.split(f)
        if tail.startswith(id):
            filename = os.path.join(directory, f)
            filenames.append(filename)
    return filenames

def copy_all(list_to_copy, destination):
    # copy_to_mydir will copy any file you give it to DST_DIR
    copy_to_mydir = partial(shutil.copy2, dst=destination)
    with ThreadPool(8) as p:
        p.map(copy_to_mydir, list_to_copy)

def prepare_train_test(directory):
    filenames = hc.find_patient_files(directory)
    num_patients = len(filenames)

    abs_patients = list()
    unk_patients = list()
    pr_patients = list()
    for i in range(num_patients):
        current_patient_data = hc.load_patient_data(filenames[i])
        murmur = hc.get_murmur(current_patient_data)
        if murmur == 'Present':
            pr_patients.append(filenames[i])
        if murmur == 'Unknown':
            unk_patients.append(filenames[i])
        if murmur == 'Absent':
            abs_patients.append(filenames[i])

    pr_train, pr_test = sklearn.model_selection.train_test_split(pr_patients, test_size=0.2)
    unk_train, unk_test = sklearn.model_selection.train_test_split(unk_patients,  test_size=0.2)
    ab_train, ab_test = sklearn.model_selection.train_test_split(abs_patients, test_size=0.2)
    files_for_train = list()
    files_for_test = list()

    for file in pr_train+unk_train+ab_train:
        head, tail = ntpath.split(file)
        id = tail.split(".")[0]
        files_for_train += find_files(directory, id)

    for file in pr_test + unk_test + ab_test:
        head, tail = ntpath.split(file)
        id = tail.split(".")[0]
        files_for_test += find_files(directory, id)


    os.mkdir("{}/{}".format(directory, "train"))
    os.mkdir("{}/{}".format(directory, "test"))
    train_files = list()
    test_files = list()
    copy_all(files_for_train, "{}/{}".format(directory, "train"))
    copy_all(files_for_test, "{}/{}".format(directory, "test"))
    return "Training data written to: {}/{}".format(directory, "train"), "Test data written to {}/{}".format(directory, "test")

prepare_train_test("/home/sfestag/Schreibtisch/physionet22/data/all")